/*
 * This file is part of source for dviviewer, 
 * a DVI file viewer for Qtopia environment. 
 * 
 * Copyright (C) 2003  hn43o.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "dvi.h"

Dvi::Dvi() : BinaryFile() {
}

Dvi::~Dvi() {
}

bool Dvi::open( QString fileName ) {
  BinaryFile::open( fileName );
  readInfo();
  return file.isOpen();
}

#define CODE_PRE 247
#define CODE_POST 248
#define CODE_POST_POST 249
#define CODE_BOP 139
#define CODE_EOP 140
#define DVIVERSION 2
#define PTEX_DVIVERSION 3
#define CODE_PADDING 223

#define PRE_SIZE ( 15 + comment.length() )
#define BOP_SIZE 45

void Dvi::readInfo() {
  if ( !file.isOpen() )
    return;

  unsigned char code, c0;

  ////////////////////
  // check PRE
  file.at( 0 );
  if ( file.size() < 15 ) {
    qDebug( "Dvi: file too short to contain PRE" );
    close();
    return;
  }
  code = getUInt1();
  if ( code != CODE_PRE ) {
    qDebug( "Dvi: file doesn't start with PRE=247" );
    close();
    return;
  }
  c0 = getUInt1();
  if ( c0 != DVIVERSION ) {
    qDebug( "Dvi: version %d != %d", c0, DVIVERSION );
    close();
    return;
  }
  scaleFactor = getUInt();
  designSize = getUInt();
  magnification = getUInt();

  unsigned char k = getUInt1();
  if ( ( int ) file.size() < 15 + k ) {
    qDebug( "Dvi: file too short to contain PRE comment" );
    close();
    return;
  }
  comment.truncate( 0 );
  for( int i=0; i<k; i++ ) 
    comment.append( getUInt1() );
  //  qDebug( "Dvi: dvi comment %s", comment.latin1() );

  ////////////////////
  // POST_POST
  file.at( file.size()-1 );
  int numPadding = 0;
  while( getUInt1() == CODE_PADDING ) { // terminates, since PRE exists.
    numPadding++;
    file.at( file.size() - 1 - numPadding );
  }
  //  qDebug( "Dvi: padding at tail = %d bytes", numPadding );
  if ( numPadding < 4 || numPadding > 7 ) {
    qDebug( "Dvi: invalid number( = %d ) of padding at tail", numPadding );
    close();
    return;
  }

  int pos = file.size() - 1 - numPadding;
  if ( pos -4 < ( int ) PRE_SIZE ) {
    qDebug( "Dvi: file too short to contain PRE + POST" );
    close();
    return;
  }
  file.at( pos );
  c0 = getUInt1();
  if ( c0 != DVIVERSION && c0 != PTEX_DVIVERSION ) {
    qDebug( "Dvi: wrong version %d != %d, %d", c0, DVIVERSION, PTEX_DVIVERSION );
    close();
    return;
  }
  isPTeX = ( c0 == PTEX_DVIVERSION );
  pos -= 4;
  file.at( pos );
  postPosition = getInt();

  // qDebug( "Dvi: POST at %d", postPosition );

  ////////////////////
  // POST
  if ( postPosition < PRE_SIZE ) {
    qDebug( "Dvi: POST before PRE ends" );
    close();
    return;
  }

  pos = postPosition;
  file.at( pos );
  code = getUInt1();
  if ( code != CODE_POST ) {
    qDebug( "Dvi: file doesn't have POST" );
    close();
    return;
  }
  bopPosition.clear();
  pageNumber.clear();
  uint pos_bop = getInt(); // saved for BOPs

  getUInt(); getUInt(); getUInt(); // num, den, mag
  getUInt(); getUInt(); getUInt2(); getUInt2(); // l, u, s, t

  // read FNTDEFs
  while( 1 ) {
    uchar code;
    int i0;

    i0 = 0;

    code = getUInt1();
    if ( code == CODE_POST_POST )
      break;
    switch( code ) {
    case 243: 
      i0 = getUInt1();
      break;
    case 244: 
      i0 = getUInt2();
      break;
    case 245: 
      i0 = getUInt3();
      break;
    case 246: 
      i0 = getInt();
      break;
    default:
      {
	qDebug( "Dvi: unexpected code %d", code );
	close();
	return;
      }
    }
    FontData *fd = new FontData;
    fd->c = getUInt();
    fd->s = getUInt();
    fd->d = getUInt();
    fd->a = getUInt1();
    fd->l = getUInt1();
    for( unsigned int i=0; i < fd->a + fd->l; i++ ) 
      fd->n.append( getUInt1() ); 
    fonts.insert( i0, fd );
  }

  pos = pos_bop;
  while( pos != -1 ) {
    if ( pos < ( int ) PRE_SIZE ) {
      qDebug( "Dvi: BOP( at %d ) befor PRE ends", pos );
      close();
      return;
    }
    bopPosition.prepend( pos );
    file.at( pos );
    code = getUInt1();
    if ( code != CODE_BOP ) {
      qDebug( "Dvi: page doesn't start with BOP" );
      close();
      return;
    }
    pageNumber.prepend( getInt() );

    file.at( pos + 41 );
    pos = getInt();
  }
  for( unsigned int i=0; i < bopPosition.count(); i++ ) 
    qDebug( "page %d at %d", pageNumber[i], bopPosition[i] );
  isValid = TRUE;
} 

void Dvi::drawPage( DviPainter *p, int page ) {
  if ( !isValid )
    return;

  direction = 0;
  cur.h = 0;
  cur.v = 0;
  cur.w = 0;
  cur.x = 0;
  cur.y = 0;
  cur.z = 0;
  stack.clear();
  //  fonts.clear();

  p->setGeometry( scaleFactor, designSize, magnification );
  p->clear();
  uint pos = bopPosition[page];
  pos += BOP_SIZE;
  file.at( pos );
  
  while( 1 ) {
    uchar code;
    uint ui0, ui1;
    int i0, i1;

    ui0 = 0;
    ui1 = 0;
    i0 = 0;
    i1 = 0;

    if ( file.atEnd() ) {
      qDebug( "Dvi: unexpected end of dvi file" );
      close();
      return;
    }
    code = getUInt1();
    if ( code == CODE_EOP )
      break;
    if ( code >= 128 && ( code <= 170 || code >= 235 ) ) {
      // fetch arguments ( and check code )
      // only the first argument for SPECIAL and FNTDEF.
      switch( code ) {
      case 138: case 141: case 142: case 147: case 152: case 161:
      case 166: 
	break;
      case 128: case 133: case 235: case 243: 
	i0 = getUInt1();
	break;
      case 239: case 255:
	ui0 = getUInt1();
	break;
      case 143: case 148: case 153: case 157: case 162: case 167: 
	i0 = getInt1();
	break;
      case 129: case 134: case 236: case 244: 
	i0 = getUInt2();
	break;
      case 240: 
	ui0 = getUInt2();
	break;
      case 144: case 149: case 154: case 158: case 163: case 168: 
	i0 = getInt2();
	break;
      case 130: case 135: case 237: case 245: 
	i0 = getUInt3();
	break;
      case 241: 
	ui0 = getUInt3();
	break;
      case 145: case 150: case 155: case 159: case 164: case 169: 
	i0 = getInt3();
	break;
      case 242: 
	ui0 = getUInt();
	break;
      case 131: case 136: case 146: case 151: case 156: case 160: 
      case 165: case 170: case 238: case 246: 
	i0 = getInt();
	break;
      case 132: case 137:
	i0 = getInt(); 
	i1 = getInt();
	break;
      default:
	{
	  qDebug( "Dvi: unexpected code%d", code );
	  close();
	  return;
	}
      }
    }
    if ( code < 128 ) {
      cur.h += p->drawChar( cur.h, cur.v, code );
    } else if ( code >= 171 && code <= 234 ) {
      fontNo = code - 171;
      p->setFont( fonts[fontNo]->n, fonts[fontNo]->s, fonts[fontNo]->d );
    } else {
      switch( code ) {
      case 128: case 129: case 130: case 131:
	cur.h += p->drawChar( cur.h, cur.v, i0 );
	break;
      case 132:
	p->drawBox( cur.h, cur.v, i1, i0 );
	cur.h += i1;
	break;
      case 133: case 134: case 135: case 136:
	p->drawChar( cur.h, cur.v, i0 );
	break;
      case 137:
	p->drawBox( cur.h, cur.v, i1, i0 );
	break;
      case 138: // NOP
	break;
	// 139 : BOP, 140: EOP
      case 141: // PUSH
	stack.push( cur );
	break;
      case 142: // POP
	cur = stack.pop();
	break;
      case 143: case 144: case 145: case 146:
	cur.h += i0;
	break;
      case 147:
	cur.h += cur.w;
	break;
      case 148: case 149: case 150: case 151:
	cur.h += i0;
	cur.w = i0;
	break;
      case 152:
	cur.h += cur.x;
	break;
      case 153: case 154: case 155: case 156:
	cur.h += i0;
	cur.x = i0;
	break;
      case 157: case 158: case 159: case 160:
	cur.v += i0;
	break;
      case 161:
	cur.v += cur.y;
	break;
      case 162: case 163: case 164: case 165: 
	cur.v += i0;
	cur.y = i0;
	break;
      case 166:
	cur.v += cur.z;
	break;
      case 167: case 168: case 169: case 170:
	cur.v += i0;
	cur.z = i0;
	break;
      case 235: case 236: case 237: case 238:
	fontNo = i0;
	p->setFont( fonts[fontNo]->n, fonts[fontNo]->s, fonts[fontNo]->d );
	break;
      case 239: case 240: case 241: case 242:
	{
	  QString s;
	  for( unsigned int i=0; i < ui0; i++ )
	    s.append( getUInt1() ); 
	  qDebug( "Dvi: special %s", s.latin1() );
	}
	break;
      case 243: case 244: case 245: case 246: 
	{
	  FontData *fd = new FontData;
	  fd->c = getUInt();
	  fd->s = getUInt();
	  fd->d = getUInt();
	  fd->a = getUInt1();
	  fd->l = getUInt1();
	  for( unsigned int i = 0; i < fd->a + fd->l; i++ ) 
	    fd->n.append( getUInt1() ); 
	  fonts.insert( i0, fd );
	}
	break;
      case 255:
	direction = ui0;
	break;
      }
    }
  }
}

QValueList<int> Dvi::pages() {
  if ( !isValid ) {
    QValueList<int> l;
    return l;
  } else
    return pageNumber;
}
