/*
 * This file is part of source for dviviewer, 
 * a DVI file viewer for Qtopia environment. 
 * 
 * Copyright (C) 2003  hn43o.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "dvipainter.h"

#include <qwidget.h>
#include <qpainter.h>

#include "jpfont.h"
#include "pkfont.h"

#define HOFFSET 65536 // 1 inch
#define VOFFSET 65536 

// (2^{-16} inch to ds)
#define INTODS(x) (( float ) ( x ) * 2.54e5 * ( float ) _designSize/( 65536.0 * ( float ) _scaleFactor))
// inverse
#define DSTOIN(x) (( float ) ( x ) * 65536.0 * ( float ) _scaleFactor/(2.54e5 * ( float ) _designSize))
#define INTOPIXEL(x) ( ( x ) * dpi / 65536 )

DviPainter::DviPainter() :
  _scaleFactor( 25400000 ), _designSize( 473628672 ), _magnification( 1000 ) 
{
}
DviPainter::~DviPainter() {
}
void DviPainter::setGeometry( uint scaleFactor, uint designSize, 
			      uint magnification ) {
  _scaleFactor = scaleFactor;
  _designSize = designSize;
  _magnification = magnification;
}

DviViewPainter::DviViewPainter() {
  fonts.setAutoDelete( TRUE );
}

void DviViewPainter::setGeometry( uint scaleFactor, uint designSize, 
				  uint magnification ) {
  DviPainter::setGeometry( scaleFactor, designSize, magnification );
}
void DviViewPainter::clear() {
  chars.resize(0);
  boxen.resize(0);
  currentFont = 0;
  bound.setRect( -HOFFSET, -VOFFSET, 541833, 766307 );
}
void DviViewPainter::setFont( QString fontName, 
			      //			      int scaleFactor, int designSize ) {
			      int, int) {
  if ( fontNotFound.contains( fontName ) ) {
    currentFont = 0;
    return;
  }

  //  qDebug( "DviViewPainter::setFont %s", fontName.latin1() );
  TexFont *f = fonts[fontName];
  if ( !f ) {
    if ( fontName.left( 3 ) == "min" || fontName.left( 4 ) == "goth" ) 
      f = new JpFont;
    else
      f = new PkFont;
    if ( f->openByName( fontName ) )
      fonts.insert( fontName, f );
    else
      fontNotFound.insert( fontName, TRUE );
  }
  currentFont = fonts[fontName];
}

void DviViewPainter::draw( QRect r ) {
  for(QDictIterator<TexFont> it(fonts); it.current(); ++it) 
    it.current()->setDpi(dpi);

  for(uint i = 0; i < chars.count(); i++) {
    Char c = chars[i];
    if (c.font) {
      QRect s = c.rect, t;
      t.setX( INTOPIXEL( HOFFSET + s.x() ) - 1);
      t.setY( INTOPIXEL( VOFFSET + s.y() ) - 2);
      t.setWidth( INTOPIXEL( s.width() ) + 2);
      t.setHeight( INTOPIXEL( s.height() ) + 2);
      if ( r.intersects( t ) )
	c.font->draw( p, 
		      INTOPIXEL( HOFFSET + c.x - c.font->hoff( c.code ) ), 
		      INTOPIXEL( VOFFSET + c.y - c.font->voff( c.code ) ), 
		      c.code );
    }
  }
  for(uint i = 0; i < boxen.count(); i++) {
    Box c = boxen[i];
    if ( c.w <= 0 || c.h <= 0 )
      continue;
    p->drawRect( INTOPIXEL( HOFFSET + c.x ), INTOPIXEL( VOFFSET + c.y ), 
		 INTOPIXEL( c.w ) - 1, INTOPIXEL( c.h ) - 1 );
  }
}

void DviViewPainter::drawBox( int x, int y, int w, int h ) {
  uint i = boxen.size();
  boxen.resize( i + 1 );
  boxen[i].x = DSTOIN( x );
  boxen[i].y = DSTOIN( y - h );
  boxen[i].w = DSTOIN( w );
  boxen[i].h = DSTOIN( h );

  bound |= QRect( x, y - h, w, h );
}

uint DviViewPainter::drawChar( int x, int y, uint code ) {
  if ( currentFont ) {
    uint i = chars.count();
    chars.resize( i + 1 );

    chars[i].x = DSTOIN(x);
    chars[i].y = DSTOIN(y);
    chars[i].rect = QRect( chars[i].x - currentFont->hoff( code ), 
			   chars[i].y - currentFont->voff( code ), 
			   currentFont->width( code ), 
			   currentFont->height( code ) );
    chars[i].code = code;
    chars[i].font = currentFont;
    uint w = currentFont->dx( code );

    bound |= chars[i].rect;

    return INTODS( w );
  } else {
    return 0;
    /*
    uint i = chars.count();
    chars.resize( i + 1 );
    chars[i].x = HOFFSET + DSTOIN( x );
    chars[i].y = VOFFSET + DSTOIN( y );
    chars[i].rect = QRect( chars[i].x, chars[i].y, 64, 64 ); // FIXME
    chars[i].code = code;
    chars[i].font = currentFont;
    QString s = QChar( code );
    return p->fontMetrics().width( s ) * denominator;
    */
  }
}

#if 0
QSize DviViewPainter::size() {
  return QSize( INTOPIXEL( bound.width() ), INTOPIXEL( bound.height() ) );
}
#endif

void DviViewPainter::setDpi( uint d ) {
  dpi = d;
}
void DviViewPainter::setPainter( QPainter *p_ ) {
  p = p_;
  p->setFont( QFont( p->font().family(), 10 ) );
}
void DviViewPainter::clearFonts() {
  fonts.clear();
}
