/*
 * This file is part of source for dviviewer, 
 * a DVI file viewer for Qtopia environment. 
 * 
 * Copyright (C) 2003  hn43o.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "dviviewer.h"
#include "fileopen.h"

#include <stdlib.h>

#include <qapplication.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qtimer.h>
#include <qvbox.h>

#include <qpe/config.h>
#include <qpe/qcopenvelope_qws.h>

#include "filedialog.h"

DviViewer::DviViewer() {
  setCaption( "DVI Viewer" );
  Config cfg( "DviViewer" );
  cfg.setGroup( "Main" );
  directory = cfg.readEntry( "Directory", QString( getenv( "HOME" ) ) );
  bool hideMenu = cfg.readBoolEntry( "HideMenu", false );
  int dpi = cfg.readNumEntry( "Dpi", 75 );
  int areaWidth = cfg.readNumEntry( "Width", 600 );
  int areaHeight = cfg.readNumEntry( "Height", 850 );

  setFocusPolicy( StrongFocus );
  dvi = new Dvi;
  vPainter = new DviViewPainter();

  QHBoxLayout *l = new QHBoxLayout( this );
  sv = new QScrollView( this );
  sv->setResizePolicy( QScrollView::AutoOne );
  if ( hideMenu ) {
    sv->setHScrollBarMode( QScrollView::AlwaysOff );
    sv->setVScrollBarMode( QScrollView::AlwaysOff );
  }
  l->addWidget( sv );
  dviArea = new DviArea( sv, dvi, vPainter );
  dviArea->resize( areaWidth, areaHeight );
  dviArea->setDpi( dpi );

  sv->addChild( dviArea );

  menuBox = new QVBox( this );
  menuBox->setFixedWidth( 80 );
  if ( hideMenu )
    menuBox->hide();
  l->addWidget( menuBox );
  pushButtonOpen = new QPushButton( menuBox );
  pushButtonOpen->setPixmap( QPixmap( qembed_findData( "fileopen.png" ) ) ); 

  QLabel *label;
  label = new QLabel( "dpi" , menuBox );
  //  spinBoxFont = new QSpinBox( 1, 10, 1, this );
  spinBoxFont = new QSpinBox( 60, 300, 1, menuBox );
  spinBoxFont->setValue( dpi );
  lastDpi = dpi;

  label = new QLabel( "width" , menuBox );
  spinBoxWidth = new QSpinBox( 200, 3000, 1, menuBox );
  spinBoxWidth->setValue( areaWidth );
  label = new QLabel( "height" , menuBox );
  spinBoxHeight = new QSpinBox( 200, 3000, 1, menuBox );
  spinBoxHeight->setValue( areaHeight );

  listBoxPage = new QListBox( menuBox );
  listBoxPage->setFocusPolicy( NoFocus );

  connect( listBoxPage, SIGNAL( highlighted( int ) ), this, SLOT( pageChanged( int ) ) );
  connect( pushButtonOpen, SIGNAL( clicked() ), this, SLOT( openPressed() ) );
  connect( spinBoxFont, SIGNAL( valueChanged( int ) ), this, SLOT( magChanged( int ) ) );
  connect( spinBoxWidth, SIGNAL( valueChanged( int ) ), this, SLOT( changeSize() ) );
  connect( spinBoxHeight, SIGNAL( valueChanged( int ) ), this, SLOT( changeSize() ) );
  connect( dviArea, SIGNAL( paintDone() ), this, SLOT( slotPaintDone() ) );
}

DviViewer::~DviViewer() {
  Config cfg( "DviViewer" );
  cfg.setGroup( "Main" );
  cfg.writeEntry( "Directory", directory );
  cfg.writeEntry( "HideMenu", menuBox->isHidden() );
  cfg.writeEntry( "Dpi", spinBoxFont->value() );
  cfg.writeEntry( "Width", spinBoxWidth->value() );
  cfg.writeEntry( "Height", spinBoxHeight->value() );
}

void DviViewer::setFile( QString fileName ) {
  dviArea->setUpdatesEnabled( FALSE );
  //  sv->setContentsPos( 64, 0 );
  dviArea->setUpdatesEnabled( TRUE );
  dvi->open( fileName );
  listBoxPage->clear();
  QValueList<int> pages = dvi->pages();
  for( unsigned int i=0; i < pages.count(); i++)
    listBoxPage->insertItem( QString::number( pages[i] ) );
  if ( pages.count() > 0 ) {
    QTimer::singleShot(0, this, SLOT(slotSetPageZero()));
    setFocus();
  }
}

void DviViewer::keyPressEvent( QKeyEvent *e ) {
  switch( e->key() ) {
  case Key_X:
    dviArea->setUpdatesEnabled( FALSE );
    if (menuBox->isHidden()) {
      sv->setHScrollBarMode( QScrollView::Auto );
      sv->setVScrollBarMode( QScrollView::Auto );
      menuBox->show();
    } else {
      sv->setHScrollBarMode( QScrollView::AlwaysOff );
      sv->setVScrollBarMode( QScrollView::AlwaysOff );
      menuBox->hide();
    }
    dviArea->setUpdatesEnabled( TRUE );
    break;
  case Key_O:
    QTimer::singleShot(0, this, SLOT(openPressed()));
    break;
  case Key_N:
    if ( listBoxPage->currentItem() + 1 < ( int ) listBoxPage->count() )
      listBoxPage->setSelected( listBoxPage->currentItem() + 1, TRUE );
    break;
  case Key_P:
    if ( listBoxPage->currentItem() > 0 )
      listBoxPage->setSelected( listBoxPage->currentItem() - 1, TRUE );
    break;
  case Key_Up:
    sv->scrollBy( 0, -32 );
    break;
  case Key_Down:
    sv->scrollBy( 0, 32 );
    break;
  case Key_PageUp:
    sv->scrollBy( 0, -sv->visibleHeight()+32 );
    break;
  case Key_PageDown:
    sv->scrollBy( 0, sv->visibleHeight()-32 );
    break;
  case Key_Left:
    sv->scrollBy( -32, 0 );
    break;
  case Key_Right:
    sv->scrollBy( 32, 0 );
    break;
  case Key_Home:
    sv->scrollBy( -sv->visibleWidth()+32, 0 );
    break;
  case Key_End:
    sv->scrollBy( sv->visibleWidth()-32, 0 );
    break;
  }
}

void DviViewer::pageChanged( int i ) {
  {
    QCopEnvelope env( "QPE/System", "busy()" );
  }
  dviArea->setUpdatesEnabled( FALSE );
  sv->setContentsPos( sv->contentsX(), 0 );
  dviArea->setUpdatesEnabled( TRUE );
  dviArea->setPage( i );
}

void DviViewer::openPressed() {
  dvi->close();
  bool result;

  DviFileDialog dlg( this, "dviviewer" ); 
  dlg.setFile( directory, file );
  dlg.showMaximized();

  result = dlg.exec();

  directory = dlg.directory();
  file = dlg.file();
  if ( result == QDialog::Accepted && !file.isNull() )
    setFile( directory + "/" + file );
}

void DviViewer::magChanged( int i ) {
  {
    QCopEnvelope env( "QPE/System", "busy()" );
  }
  dviArea->setUpdatesEnabled( FALSE );
  //  dviArea->setDpi( 25 * ( i + 2) );
  dviArea->setDpi( i );
  sv->setContentsPos( sv->contentsX() * i / lastDpi, 
		      sv->contentsY() * i / lastDpi );
  lastDpi = i;
  dviArea->setUpdatesEnabled( TRUE );
  dviArea->update();
  setFocus();
}

void DviViewer::changeSize() {
  dviArea->resize( spinBoxWidth->value(), spinBoxHeight->value() );
}

// hack against flicker
void DviViewer::slotSetPageZero() {
  listBoxPage->setSelected( 0, TRUE );
}

void DviViewer::slotPaintDone( ) {
  QCopEnvelope env( "QPE/System", "notBusy(QString)" );
  env << "";
}

DviArea::DviArea( QWidget *parent, Dvi *d, DviViewPainter *dp ) : 
  QWidget( parent ), dvi( d ), dviViewPainter( dp ), page( -1 ) {
  setBackgroundColor( QColor( 255, 255, 255 ) );
}

void DviArea::setDpi( uint d ) {
  dviViewPainter->setDpi( d );

  //  setUpdatesEnabled( FALSE );
  //  resize( dviViewPainter->size() );
  //  setUpdatesEnabled( TRUE );
}

void DviArea::setPage( uint i ) {
  page = i;
  dvi->drawPage( dviViewPainter, page );

  //  setUpdatesEnabled( FALSE );
  //  resize( dviViewPainter->size() );
  //  setUpdatesEnabled( TRUE );

  update();
}

void DviArea::paintEvent( QPaintEvent *e ) {
  if (page < 0)
    return;
  QPainter p( this );
  dviViewPainter->setPainter( &p );
  dviViewPainter->draw( e->rect() );

  emit paintDone();
}
