/*
 * This file is part of source for dviviewer, 
 * a DVI file viewer for Qtopia environment. 
 * 
 * Copyright (C) 2003  hn43o.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pkfont.h"

#include <qimage.h>
#include <qpainter.h>

#define FONTDIR ( "/home/QtPalmtop/share/texfonts/" )

PkFont::PkFont() {
  dpi = 0;
  data.setAutoDelete( TRUE );
}

bool PkFont::open( QString fileName ) {
  BinaryFile::open( fileName );
  readInfo();
  qDebug( "PkFont: %s %d", fileName.latin1(), file.isOpen() );
  return file.isOpen();
}

#define FILE_DPI 300
bool PkFont::openByName( QString name ) {
  return open( QString( FONTDIR ) + name + QString( ".%1pk" ).arg( FILE_DPI ) );
}

uint PkFont::dx( uint code ) { // in inch, multiplied by 65536
  PkCharData *d = data[code];
  if ( d )
    return d->dx/FILE_DPI;
  return 0;
}
uint PkFont::width( uint code ) { // in inch, multiplied by 65536
  PkCharData *d = data[code];
  if ( d )
    return ( d->w << 16 )/FILE_DPI;
  return 0;
}
uint PkFont::height( uint code ) { // in inch, multiplied by 65536
  PkCharData *d = data[code];
  if ( d )
    return ( d->h << 16 )/FILE_DPI;
  return 0;
}
int PkFont::hoff( uint code ) { // in inch, multiplied by 65536
  PkCharData *d = data[code];
  if ( d )
    return ( int ) ( ( d->hoff * 65536.0 )/FILE_DPI );
  return 0;
}
int PkFont::voff( uint code ) { // in inch, multiplied by 65536
  PkCharData *d = data[code];
  if ( d )
    return ( int ) ( ( d->voff * 65536.0 )/FILE_DPI );
  return 0;
}
uint PkFont::tfm( uint code ) { 
  PkCharData *d = data[code];
  if ( d )
    return d->tfm;
  return 0;
}
uint PkFont::designSize() {
  return _designSize;
}

void PkFont::setDpi( uint d ) {
  if ( dpi == d )
    return;
  dpi = d;
  for(QIntDictIterator<PkCharData> it(data); it.current(); ++it)
    it.current()->pm.resize(0, 0);
}

void PkFont::draw( QPainter *painter, int x, int y, uint code ) {
  //  qDebug( "Rendering %c", code );
  PkCharData *d = data[code];
  if ( !d )
    return;
  if ( d->pm.isNull() ) {
    QPixmap pm;
    pm.resize( d->w, d->h );
    QPainter p( &pm );

    file.at( d->dataPos );
    //    qDebug( "data position: %x", d->dataPos );

    if ( d->dyn_f == 14 ) {
      uchar c = getUInt1();
      int ptr = 7;
      for( unsigned int y = 0; y < d->h; y++ )
	for( unsigned int x = 0; x < d->w; x++ ) {
	  if ( c & ( 1<<ptr ) )
	    p.setPen( QColor( 0, 0, 0 ) );
	  else
	    p.setPen( QColor( 255, 255, 255 ) );
	  p.drawPoint( x, y );
	  ptr--;
	  if ( ptr < 0 ) {
	    c = getUInt1();
	    ptr = 7;
	  }
	}
    } else {
      secondNybble = FALSE;
      bool black = d->black;
      bool b[d->h][d->w];
      unsigned int x = 0, y = 0;
      int rep_y = -1;
      int rc = 0;
      while( y < d->h ) {
	repeatCount = 0;
	int a = decodePackedNum( d->dyn_f );
	if ( repeatCount > 0 ) 
	  rc = repeatCount;
	//      qDebug( "x=%d, y=%d, %d %d %d", x, y, a, repeatCount, rc );
	for( int j=0; j < a; j++ ) {
	  b[y][x] = black ? 1 : 0;
	  x++;
	  if ( x >= d->w ) {
	    x = 0;
	    if ( repeatCount > 0 ) 
	      rep_y = y;
	    y++;
	    if ( rc > 0 && rep_y != ( int ) y )
	      for( unsigned int l = 0; ( int ) l < rc; l++ ) {
		for( unsigned int k = 0; k < d->w; k++ ) {
		  b[y][k] = b[y-1][k];
		}
		y++;
	      }
	    rc = 0;
	  }
	}
	black = !black;
      }
      for( unsigned int y = 0; y < d->h; y++ )
	for( unsigned int x = 0; x < d->w; x++ ) {
	  if ( b[y][x] ) 
	    p.setPen( QColor( 0, 0, 0 ) );
	  else
	    p.setPen( QColor( 255, 255, 255 ) );
	  p.drawPoint( x, y );
	}
    }
    if ( !pm.isNull() ) {
      //	  uint w = QMAX( ( d->w*dpi+FILE_DPI/2 )/FILE_DPI, 1 ), 
      //		h = QMAX( ( d->h*dpi+FILE_DPI/2 )/FILE_DPI, 1 );
      uint w = QMAX( d->w*dpi/FILE_DPI, 1 ), 
	h = QMAX( d->h*dpi/FILE_DPI, 1 );
      d->pm.convertFromImage( pm.convertToImage().smoothScale( w, h ) );
    }
  }
  painter->drawPixmap( x, y, d->pm );
}

#define CODE_POST 245
#define CODE_NOP 246
#define CODE_PRE 247

#define PKVERSION 89

void PkFont::readInfo() {
  if ( !file.isOpen() )
    return;
  uchar code, c0;

  file.at( 0 );
  if ( file.size() < 3 ) { // until 'k'.
    qDebug( "PkFont: file too short to contain PRE" );
    close();
    return;
  }
  code = getUInt1();
  if ( code != CODE_PRE ) {
    qDebug( "PkFont: file doesn't start with PRE" );
    close();
    return;
  }
  c0 = getUInt1();
  if ( c0 != PKVERSION ) {
    qDebug( "PkFont: version %d != %d", c0, PKVERSION );
    close();
    return;
  }
  uchar k = getUInt1();
  if ( ( int ) file.size() < 19 + k ) {
    qDebug( "PkFont: file too short to contain PRE comment" );
    close();
    return;
  }
  QString comment;
  for( int i=0; i<k; i++ ) 
    comment.append( getUInt1() );

  _designSize = getUInt();
  checkSum = getUInt();
  hppp = getUInt();
  vppp = getUInt();
  qDebug( "Pk: comment '%s' ds=%d hppp=%d vppp=%d", comment.latin1(), 
	  _designSize, hppp, vppp );

  secondNybble = FALSE;

  // read per-character data
  uint pos = file.at();
  while( 1 ) {
    uchar type; // 0: short, 1: ext short, 2: long

    PkCharData *d = new PkCharData;

    //	qDebug( "Pk: pos = %d", pos );
    file.at( pos );

    uchar flagByte = getUInt1();
    //	  qDebug( "%d", flagByte );
    if ( flagByte == CODE_POST )
      break;
    if ( flagByte >= 240 ) {
      switch( flagByte ) {
      case 240: 
	pos += 2 + getUInt1();
	break;
      case 241: 
	pos += 3 + getUInt2();
	break;
      case 242: 
	pos += 4 + getUInt3();
	break;
      case 243: 
	pos += 5 + getUInt();
	break;
      case 244: 
	pos += 5;
	break;
      case 246: 
	pos += 1;
	break;
      }
      continue;
    }

    d->dyn_f = flagByte >> 4;
    d->black = ( ( flagByte & 8 ) != 0 );
    if ( ( flagByte & 7 ) == 7 ) {
      type = 2;
      d->dataPos = pos + 37;
    } else {
      if ( flagByte & 4 ) {
	type = 1;
	d->dataPos = pos + 17;
      } else {
	type = 0;
	d->dataPos = pos + 11;
      }
    }

    uint packetLength;
    uint code;
    switch( type ) {
    case 0:
      packetLength = ( ( flagByte & 3 ) << 8 ) + getUInt1();
      code = getUInt1();
      d->tfm = getUInt3();
      d->dx = getUInt1() << 16;
      d->dy = 0;
      d->w = getUInt1();
      d->h = getUInt1();
      d->hoff = getInt1();
      d->voff = getInt1();
      pos += 3 + packetLength;
      break;
    case 1:
      packetLength = ( ( flagByte & 3 ) << 16 ) + getUInt2();
      code = getUInt1();
      d->tfm = getUInt3();
      d->dx = getUInt2() << 16;
      d->dy = 0;
      d->w = getUInt2();
      d->h = getUInt2();
      d->hoff = getInt2();
      d->voff = getInt2();
      pos += 4 + packetLength;
      break;
    case 2:
      packetLength = getUInt();
      code = getUInt();
      d->tfm = getUInt();
      d->dx = getUInt();
      d->dy = getUInt();
      d->w = getUInt();
      d->h = getUInt();
      d->hoff = getInt();
      d->voff = getInt();
      pos += 9 + packetLength;
      break;
    }
    //	qDebug( "Pk: %d type = %d, w = %d h = %d, hoff=%d", code, type, d->w, d->h, d->hoff );
    data.insert( code, d );
  }
}

uchar PkFont::getNybble() {
  if ( secondNybble ) {
    secondNybble = FALSE;
    //	qDebug( " %x", currentByte & 0xf );
    return currentByte & 0xf;
  } else {
    currentByte = getUInt1();
    secondNybble = TRUE;
    //	qDebug( " %x", currentByte >> 4 );
    return currentByte >> 4;
  }
}

// from dvistd0.tex
uint PkFont::decodePackedNum( uchar dyn_f ) {
  uint i, j;
  i = getNybble();
  if ( i == 0 ) {
    while( 1 ) {
      j = getNybble(); 
      i++;
      if ( j!=0 ) 
	break;
    }
    while( i > 0 ) {
      j = ( j<<4 ) + getNybble();
      i--;
    }
    return j - 15 + ( 13 - dyn_f ) * 16 + dyn_f;
  } else {
    if ( i <= dyn_f )
      return i;
    else if ( i < 14 ) {
      return ( i - dyn_f - 1 )*16 + getNybble() + dyn_f + 1;
    } else {
      if ( repeatCount ) {
	qDebug( "PkFont: second repeat count for this row" );
	return 0;
      } 
      repeatCount = 1;
      if ( i == 14 )
	repeatCount = decodePackedNum( dyn_f );
      return decodePackedNum( dyn_f );
    }
  }
}

